/* jshint esversion: 6 */
/* eslint no-console: "off" */

module.exports.process = mdcontent => (new MdDom(mdcontent)).getRoot();

const DEBUG_DUMP = false;

function log() {
    if (DEBUG_DUMP) {
        for (var i = 0; i < arguments.length; i++) {
            var o = arguments[i];
            o = typeof o == "object" ? JSON.stringify(o, null, 4) : o;
            console.log(o);
        }
    }
}

class MdDom {
    constructor(mdstring) {
        this.root = { children: [] };
        this.nesting = [this.root];
        this.lastItemType = null;
        if (mdstring) { this.parseString(mdstring); }
    }
    add(obj) {
        if (obj && typeof obj.level === "number") {
            if (obj.type == "BLANK") {
                log("IGNORE BLANK");
            } else {
                log("ADD ITEM", obj);
                let parent = this.parentByObject(obj);
                if (parent && parent.children) {
                    // associate objects
                    parent.children.push(obj);
                    // manage nesting
                    if (obj.level > 0) {
                        this.nesting.splice(obj.level);
                        this.nesting[obj.level] = obj;
                    }
                }
                // manage bullet groups in the parent (unbroken runs of bullet items)
                if (obj.type == "BULLET" && parent && parent.bulletGroups) {
                    if (this.lastItemType != obj.type) {
                        parent.bulletGroups.push([]);
                        log("New Bullet Group");
                    }
                    parent.bulletGroups[parent.bulletGroups.length - 1].push(obj);
                }
                // manage quote groups in the parent (unbroken runs of quote items)
                if (obj.type == "QUOTE" && parent && parent.quoteGroups) {
                    if (this.lastItemType != obj.type) {
                        log("New Quote Group");
                        parent.quoteGroups.push([]);
                    }
                    parent.quoteGroups[parent.quoteGroups.length - 1].push(obj);
                }
            }
            this.lastItemType = obj.type;
        }
    }
    get(level) {
        return this.nesting[level];
    }
    getRoot() {
        return this.root;
    }
    parentByObject(obj) {
        return this.parentByLevel(obj.level);
    }
    parentByLevel(level) {
        if (level == 0) {
            return this.nesting.reduce((a, b) => b || a);
        } else {
            return this.nesting[level - 1] || this.nesting.reduce((a, b, i) => (i < level && b ? b : a), null);
        }
    }
    parseString(str) {
        this.parseLines(str.replace(/\r/g, "") // remove carriage returns
            .split("\n") // split on line breaks
            //.filter(l => l && l.length > 0) // retain only non-blank lines
        );
    }
    parseLines(lines) {
        lines.forEach(line => this.add(new MdItem(line)));
    }
}

class MdItem {
    constructor(mdLine) {
        this.raw = mdLine; //
        this.ws_tabs = null; //
        this.ws_spaces = null; //
        this.indent = 0; //
        this.md = null; //
        this.type = null; //
        this.level = 0; //
        this.rol = null; //
        this.tokens = []; //
        this.strings = []; //
        this.codestrings = []; //
        this.keys = {}; //
        this.key = null; //
        this.options = {}; //
        // these collections managed by MdDom
        this.children = [];
        this.bulletGroups = [];
        this.quoteGroups = [];

        if (!mdLine || mdLine.match(/^\s*$/)) {
            this.type = "BLANK";
        } else {
            [, this.ws_tabs = "", this.ws_spaces = "", this.md = "", this.rol = ""] = mdLine.match(/^(\t*)((?: {2})*)(#{1,5}|-|>)?\s*(.*?)\s*$/);

            let rx = /(['"`])(?:[^\1\\]*?(?:\\.[^\1\\]*?)*?)\1|\S+/g, token, tokens, quoted, key, value, option;

            //while (match = rx.exec(this.rol)) { tokens.push(match[0]); }
            tokens = this.rol.match(rx) || [];

            // Iterate in reverse so that key values are already processed (e.g. unquoted)
            for (var i = tokens.length - 1; i >= 0; i--) {
                token = tokens[i];
                quoted = token.match(/^(["'`])(.*)\1$/);
                key = token.match(/^(.+):$/);
                option = token.match(/^([^=]+)=([^=]*)$/);
                if (quoted) {
                    token = quoted[2];
                    (quoted[1] == "`" ? this.codestrings : this.strings).unshift(token);
                } else if (key) {
                    key = key[1];
                    this.key = this.key || key;
                    this.keys[key] = this.tokens[i + 1];
                } else if (option) {
                    [, key, value] = option;
                    this.options[key] = decodeURIComponent(value);
                }
                this.tokens[i] = token;
            }

            this.level = this.md.match(/^(#*)/)[1].length;
            this.indent = this.ws_tabs + (this.ws_spaces / 2);
            this.type = ({ "#": "HEADING", "-": "BULLET", ">": "QUOTE" })[this.md[0]] || (this.codestrings.length == this.tokens.length ? "CODE" : this.codestrings.length == 0 ? "TEXT" : "MIXED");
        }
    }
}
