#!/usr/bin/env node

/* jshint esversion: 6 */

// ---------------------------------------------------------------------
// Command line params and defaults

const outputFolderDefault = './';
const plexiformconfigname = 'plexiformconfig.json';

let galleryUri = "https://plexiform-gallery.azurewebsites.net";

// ---------------------------------------------------------------------
// Require path

const path = require('path');
const fs = require('fs-extra');
const fetch = require('node-fetch');

const processmd = require("./includes/process-md").process;

/* The spec, build and templates are all 'require'd, and will be relative to your cwd, so push that (and node_modules sub-dir) onto our module paths list */
module.paths.push(process.cwd());
module.paths.push(path.normalize(`${process.cwd()}/node_modules`));

// ---------------------------------------------------------------------
// App info

var app = {
    name: "Plexiform",
    version: "3.5.1",
    author: "Raith"
};

var appSignature = `${app.name} v${app.version} by ${app.author}`;

// ---------------------------------------------------------------------
// Initialise Utils and Spec

var base_utils = require("./includes/base-utils.js").utils;

base_utils.appinfo = app;

var base_spec = require("./includes/base-spec.json");

var augmentSpec = cfg => Object.assign(cfg, { app: app }, base_spec);

// ---------------------------------------------------------------------
// Process args

var recognisedCommands = { "init": true, "build": true, "gallery": true };

var argv = require('process-argv')();
argv.command = argv.command || null;
argv.options = argv.options || {};
argv.params = argv.params || [];

// switches
var opt_help = "help" in argv.options;
var opt_force_overwrite = "force" in argv.options;

// parameterised options
var spec_name = argv.options.spec || null;
var build_name = argv.options.build || "*";
var output_folder = argv.options.output || outputFolderDefault;

if ("debug" in argv.options) {
    console.log("DEBUG INFO:\n");
    console.log(`build_name = ${build_name}`);
    if (build_name) console.log(`    resolved = ${require.resolve(build_name)}`);
    console.log(`spec_name = ${spec_name}`);
    if (spec_name) console.log(`    resolved = ${require.resolve(spec_name)}`);
    console.log(`output_folder = ${output_folder}`);
    console.log("Module paths:", module.paths);
    console.log("Require paths:", require.paths);
    console.log("Current working directory:", process.cwd());
}

if (argv.command && !(argv.command in recognisedCommands)) {
    console.log(`ERROR! Unrecognised command ${argv.command}`);
}

if ("version" in argv.options) {
    console.log(app.version);
    process.exit();
}

if ("help" in argv.options || !(argv.command in recognisedCommands)) {
    console.log(`
${appSignature}

usage:

plexiform --help
plexiform --version

plexiform init [--force]
plexiform init [--force] [--spec <spec_name>] [--build <build_name>] [--output <output_folder>]

plexiform build [--save]
plexiform build --spec <spec_name> [--build <build_name>] [--output <output_folder>] [--save]

plexiform gallery list
plexiform gallery query     --spec <spec_name> --secret <secret>
plexiform gallery new       --spec <spec_name> --secret <secret> --description <description>
plexiform gallery secret    --spec <spec_name> --secret <secret> --newsecret <new_secret>
plexiform gallery rename    --spec <spec_name> --secret <secret> --newname <new_name>
plexiform gallery describe  --spec <spec_name> --secret <secret> --newdescription <new_description>
plexiform gallery save      --spec <spec_name> --secret <secret> --semver <sementic_version> --author <author_name> --details <details> [--file <file_path>]
plexiform gallery publish   --spec <spec_name> --secret <secret> [--rev <revision_number>]
plexiform gallery fetch     --spec <spec_name> --secret <secret> [--rev <revision_number>]

All gallery commands also support the optional --gallery <gallery_URI> switch for specifying the gallery URI

===============================================================================

"plexiform init"
Creates a ${plexiformconfigname} file. For subsequent use without needing parameters.

"plexiform build" without the spec_name being specified will require the ${plexiformconfigname} file in the CURRENT WORKING DIRECTORY.
In this case, the specified spec_name, build_name, and output_folder are as-per the parameters explained hereafter, including that they should be relative to the cwd, or elsewhere in the 'module paths' (where they can be require'd).

<spec_name>
This should refer to a "json", "md", or other file, relative to the CWD.
It will expose a specification object, either directly (json) or via the build's "process" method.
If using md then it will be parsed into plexiform's proprietary MdDom object prior to passing into the process method.
Any other filetypes are passed in raw.

NOTE: If the --save switch is applied then the processed spec is stringified to a file with the original spec_name plus the ".json" extension, with FORCED overwriting!

<build_name>
This should refer to a file with extension ".js" in the CWD or modules path (for require).
It is a node module which exports a method named "build" with signature (spec, utils) => void which will perform your build.
It may also export a method named "process" with signature (input) => specification_object which will process a non-json specification file.
If no <build_name> is specified then "*" is assumed, unless it is looked up from ${plexiformconfigname}
If "*" is specified then all modules named "plexiform-build-*" are required from the current node_modules path.

<output_folder>
The folder into which files are generated.
If not specified then defaults to "${outputFolderDefault}"

Mixing build syntax
-------------------
The plexiform config file will only be loaded if no build_name is specified.
If the config file is loaded then any other specified option will be overridden - the file has precedence.

Template files
--------------
These should reside in ./templates/ (relative to your build script) with extension ".js"
These are node modules which export a single method named "template" with signature (context, spec, utils): string which will return a templated output.
These should be consumed as necessary by your build script.

`);
    process.exit();
}

var pconfig = null, spec, build;

if (argv.command == "init") {
    if (!fs.existsSync(plexiformconfigname) || opt_force_overwrite) {
        build_name = build_name || "*";
        spec_name = spec_name || "./path/to/my_app_spec.json|md";
        let configcontent = `{
    "build_name": "${build_name}",
    "spec_name": "${spec_name}",
    "output_folder": "${output_folder}"
}`;
        fs.outputFileSync(plexiformconfigname, configcontent);
    } else {
        console.log("ERROR! Config file exists. To overwrite it use the --force option.");
    }
}

function expandBuilds() {
    var builds = [];
    if (build_name == "*") {
        console.log(`Finding "plexiform-build-*" scripts`);
        var searchpath = `${process.cwd()}/node_modules`;
        if (fs.existsSync(searchpath)) {
            var files = fs.readdirSync(searchpath);
            for (var i = 0; i < files.length; i++) {
                var filename = files[i];
                var filepath = path.join(searchpath, filename);
                var stat = fs.lstatSync(filepath);
                if ((stat.isDirectory() || stat.isSymbolicLink())&& filename.match(/^plexiform-build-/)) {
                    console.log(`Build script: ${filename}`);
                    builds.push(filename);
                }
            }
        }
    } else {
        builds.push(build_name);
    }
    return builds;
}

if (argv.command == "build") {

    var configs = null;

    if (!build_name || !spec_name) {
        try {
            console.log(`Looking for plexiform config "${plexiformconfigname}"`);
            pconfig = require(plexiformconfigname);
            if (pconfig !== null) {
                if (Array.isArray(pconfig)) {
                    configs = pconfig;
                } else {
                    ({ build_name, spec_name, output_folder } = pconfig);
                }
                console.log("Loaded");
            }
        } catch (e) {
            console.log(`ERROR! Cannot find ${plexiformconfigname}. See --init usage.`);
        }
    }

    if (!configs) configs = [{ build_name, spec_name, output_folder }];

    for (var i = 0; i < configs.length; i++) {
        ({ build_name, spec_name, output_folder } = configs[i]);

        var builds = expandBuilds();

        if (builds.length > 0 && spec_name) {
            base_utils.output_folder = output_folder;
        
            var rawspec, specobj;
            try {
                if (spec_name.match(/\.(json|js)$/)) {
                    spec = require(spec_name);
                } else {
                    rawspec = fs.readFileSync(spec_name).toString();

                    // Process .md file into MdDom object, otherwise leave as raw content
                    specobj = (spec_name.match(/\.md$/)) ? processmd(rawspec) : rawspec;
                }
            } catch (e) {
                console.log(`ERROR! Cannot find spec file "${spec_name}" to require.`);
            }

            if (spec || specobj) {
                // ensure we have an output folder
                if (!fs.existsSync(output_folder)) fs.mkdirSync(output_folder);

                builds.forEach(build_name => {

                    console.log("\nBUILDING...\n");
                    console.log(`${appSignature}\n\nbuild_name: ${build_name}\nspec_name: ${spec_name}\noutput_folder: ${output_folder}`);

                    if (specobj) {
                        // Process the spec as necessary
                        try {
                            var processor = require(build_name).process;
                            spec = processor(specobj);
                            console.log(`Processed spec using ${build_name}.process`);

                            if ("save" in argv.options) {
                                var specSaveName = spec_name + '.json';
                                fs.writeFileSync(specSaveName, JSON.stringify(spec, null, 4));
                            }
                        } catch (e2) {
                            console.log(`ERROR! Not able to process spec using ${build_name}.process"`);
                        }
                    }

                    // add the app and base spec to our loaded/processed spec
                    augmentSpec(spec);

                    try {
                        build = require(build_name).build;
                    } catch (e) {
                        console.log(`ERROR! Cannot require build file ${build_name}.`);
                    }

                    try {
                        build(spec, base_utils);
                    } catch (e) {
                        console.log(`ERROR! Cannot find execute ${build_name}.build.`);
                    }
                
                });

            }
        
        } else {
            console.log(`ERROR! Not enough parameters, or unable to load ${plexiformconfigname}`);
        }
    }
    
}

if (argv.command == "gallery") {
    console.log(`Plexiform Gallery - ${argv.params[0] || "please specify a command"}`);

    var overrideGalleryUri = argv.options["gallery"];
    if (overrideGalleryUri) {
        if (!overrideGalleryUri.match(/^https?:\/\//)) overrideGalleryUri = `http://${overrideGalleryUri}`;
        console.log(`Using gallery URI = ${overrideGalleryUri}`);
        galleryUri = overrideGalleryUri;
    }

    // plexiform gallery publish --spec <spec_name> --secret <secret> [--rev <revision_number>]
    if (argv.params[0] == "publish") {
        var name = argv.options.spec || null;
        var secret = argv.options.secret || null;
        var revisionNumber = argv.options.rev || null;
        if (name && secret) {
            var payload = {
                specname: name,
                secret: secret,
                revisionNumber: revisionNumber
            };
            fetch(`${galleryUri}/api/cmd/Publishing/Publish`, fetch_options(payload))
                .then(res => res.json()).then(o => {
                    if (o.isSuccess) {
                        var spec = o.value.spec;
                        if (spec) {
                            console.log(`Published Revision #${spec.currentRevision} for Specification "${spec.name}"`);
                        } else {
                            console.log(`Not found Specification "${name}", or incorrect secret.`);
                        }
                    } else {
                        dump(o);
                    }
                })
                .catch(exception_handler);
        } else {
            console.log("Invalid options.");
            console.log("Usage: plexiform gallery secret  --spec <spec_name> --secret <secret> --newsecret <new_secret>");
        }
    }

    // plexiform gallery secret  --spec <spec_name> --secret <secret> --newsecret <new_secret>
    if (argv.params[0] == "secret") {
        var name = argv.options.spec || null;
        var secret = argv.options.secret || null;
        var newsecret = argv.options.newsecret || null;
        if (name && secret && newsecret) {
            var payload = {
                specname: name,
                secret: secret,
                newsecret: newsecret
            };
            fetch(`${galleryUri}/api/cmd/Publishing/Secret`, fetch_options(payload))
                .then(res => res.json()).then(o => {
                    if (o.isSuccess) {
                        if (o.value.result) {
                            console.log(`Changed secret for Specification "${name}"`);
                        } else {
                            console.log(`Not found Specification "${name}", or incorrect secret.`);
                        }
                    } else {
                        dump(o);
                    }
                })
                .catch(exception_handler);
        } else {
            console.log("Invalid options.");
            console.log("Usage: plexiform gallery secret  --spec <spec_name> --secret <secret> --newsecret <new_secret>");
        }
    }
    
    // plexiform gallery rename  --spec <spec_name> --secret <secret> --newname <new_name>
    if (argv.params[0] == "rename") {
        var name = argv.options.spec || null;
        var secret = argv.options.secret || null;
        var newname = argv.options.newname || null;
        if (name && secret && newname) {
            var payload = {
                specname: name,
                secret: secret,
                newname: newname
            };
            fetch(`${galleryUri}/api/cmd/Publishing/Rename`, fetch_options(payload))
                .then(res => res.json()).then(o => {
                    if (o.isSuccess) {
                        if (o.value.result) {
                            console.log(`Changed name of Specification "${name}"`);
                        } else {
                            console.log(`Not found Specification "${name}", or incorrect secret.`);
                        }
                    } else {
                        dump(o);
                    }
                })
                .catch(exception_handler);
        } else {
            console.log("Invalid options.");
            console.log("Usage: plexiform gallery rename  --spec <spec_name> --secret <secret> --newname <new_name>");
        }
    }

    // plexiform gallery describe  --spec <spec_name> --secret <secret> --newdescription <new_description>
    if (argv.params[0] == "describe") {
        var name = argv.options.spec || null;
        var secret = argv.options.secret || null;
        var newdescription = argv.options.newdescription || null;
        if (name && secret && newdescription) {
            var payload = {
                specname: name,
                secret: secret,
                newdescription: newdescription
            };
            fetch(`${galleryUri}/api/cmd/Publishing/Describe`, fetch_options(payload))
                .then(res => res.json()).then(o => {
                    if (o.isSuccess) {
                        if (o.value.result) {
                            console.log(`Changed description for Specification "${name}"`);
                        } else {
                            console.log(`Not found Specification "${name}", or incorrect secret.`);
                        }
                    } else {
                        dump(o);
                    }
                })
                .catch(exception_handler);
        } else {
            console.log("Invalid options.");
            console.log("Usage: plexiform gallery describe  --spec <spec_name> --secret <secret> --newdescription <new_description>");
        }
    }

    // plexiform gallery fetch --spec <spec_name> --secret <secret> [--rev <revision_number>]
    if (argv.params[0] == "fetch") {
        var name = argv.options.spec || null;
        var secret = argv.options.secret || null;
        var revisionNumber = argv.options.rev || null;
        if (name && secret) {
            fetch(`${galleryUri}/api/qry/Publishing/get/${name}/${secret}/${revisionNumber}`, fetch_options())
                .then(res => res.json()).then(o => {
                    if (o.isSuccess) {
                        var rev = o.value[0];
                        if (rev) {
                            fs.writeFileSync(`./${name}`, rev.content);
                            console.log(`Fetched Revision #${rev.revisionNumber} of Specification "${rev.spec.name}"`);
                        } else {
                            console.log(`Not found Revision #${revisionNumber == null ? "current" : revisionNumber} for Specification "${name}", or incorrect secret.`);
                        }
                    } else {
                        dump(o);
                    }
                })
                .catch(exception_handler);
        } else {
            console.log("Invalid options.");
            console.log("Usage: plexiform gallery fetch --spec <spec_name> --secret <secret> [--rev <revision_number>]");
        }
    }

    // plexiform gallery save --spec <spec_name> --secret <secret> --semver <sementic_version> --author <author_name> --details <details> [--file <file_path>]
    if (argv.params[0] == "save") {
        var name = argv.options.spec || null;
        var secret = argv.options.secret || null;
        var semver = argv.options.semver || "?.?.?";
        var details = argv.options.details || "";
        var author = argv.options.author || "unknown";
        var file_path = argv.options.file || `./${name}`;
        if (name && secret) {
            var file_content = fs.readFileSync(file_path).toString();
            var revObj = {
                ID: 0,
                revisionNumber: 0,
                revisionDate: (new Date()).toISOString(),
                revisionDetails: details,
                semver: semver,
                author: author,
                content: file_content
            };
            var payload = {
                specname: name,
                secret: secret,
                rev: revObj
            };
            fetch(`${galleryUri}/api/cmd/Publishing/Save`, fetch_options(payload))
                .then(res => res.json()).then(o => {
                    if (o.isSuccess) {
                        var rev = o.value.rev;
                        if (rev) {
                            console.log(`Added new Revision #${rev.revisionNumber} with ID [${rev.ID}]`);
                        } else {
                            console.log(`Specification "${name}" not found, or incorrect secret.`);
                        }
                    } else {
                        dump(o);
                    }
                })
                .catch(exception_handler);
        } else {
            console.log("Invalid options.");
            console.log("Usage: plexiform gallery save --spec <spec_name> --secret <secret> --semver <sementic_version> --author <author_name> --details <details> [--file <file_path>]");
        }
    }
    
    // plexiform gallery query --spec <spec_name> --secret <secret>
    if (argv.params[0] == "query") {
        var name = argv.options.spec || null;
        var secret = argv.options.secret || null;
        if (name && secret) {
            fetch(`${galleryUri}/api/qry/Publishing/query/${name}/${secret}`, fetch_options())
                .then(res => res.json()).then(o => {
                    if (o.isSuccess) {
                        var spec = o.value[0];
                        if (spec) {
                            console.log(`${spec.labelL}\n\n${spec.revisions.length} Revision${spec.revisions.length == 1 ? "" : "s"}:`);
                            for (var i = 0; i < spec.revisions.length; i++) {
                                console.log(spec.revisions[i].labelL);
                            }
                        } else {
                            console.log(`Specification "${name}" not found, or incorrect secret.`);
                        }
                    } else {
                        dump(o);
                    }
                })
                .catch(exception_handler);
        } else {
            console.log("Invalid options.");
            console.log("Usage: plexiform gallery query --spec <spec_name> --secret <secret>");
        }
    }

    // plexiform gallery new --spec <spec_name> --secret <secret> --description <description>
    if (argv.params[0] == "new") {
        var name = argv.options.spec || null;
        var secret = argv.options.secret || null;
        var description = argv.options.description || "";
        if (name && secret && description) {
            var specObj = {
                ID: 0,
                name: name,
                secret: secret,
                description: description,
                logo: "",
                currentRevision: 0
            };
            var payload = {
                spec: specObj
            };
            fetch(`${galleryUri}/api/cmd/Publishing/New`, fetch_options(payload))
                .then(res => res.json()).then(o => {
                    if (o.isSuccess) {
                        var spec = o.value.spec;
                        console.log(`Added new Specification "${spec.name}" with ID [${spec.ID}]`);
                    } else {
                        dump(o);
                    }
                })
                .catch(exception_handler);
        } else {
            console.log("Invalid options.");
            console.log("Usage: plexiform gallery new --spec <spec_name> --secret <secret> --description <description>");
        }
    }

    // plexiform gallery list
    if (argv.params[0] == "list") {
        fetch(`${galleryUri}/api/qry/Publishing/list`, fetch_options())
            .then(res => res.json()).then(o => {
                if (o && o.isSuccess) {
                    var data = o.value;
                    console.log(`Gallery has ${data.length} specification${data.length == 1 ? "" : "s"}:`);
                    for (var i = 0; i < data.length; i++) {
                        var spec = data[i];
                        console.log(spec.labelL);
                    }
                }
            })
            .catch(exception_handler);
    }

}

function fetch_options(postdata = null, method = "GET") {
    var opts = {
        method: (postdata != null ? "POST" : method),
        headers: {
            "Accept": "application/json, text/plain, */*",
            "Content-Type": "application/json"
        }
    };
    if (postdata) opts.body = JSON.stringify(postdata);
    return opts;
}

function exception_handler(e) {
    console.log("Oh dear, something went wrong! :(");
    console.log(e);
}

function dump() {
    for (var i = 0; i < arguments.length; i++) {
        console.log(JSON.stringify(arguments[i], null, 4));
    }
}

// ---------------------------------------------------------------------
